#ifndef ZAD2_QUADRUPLEGENERATOR_H
#define ZAD2_QUADRUPLEGENERATOR_H


#include <filesystem>
#include <utility>
#include "../TypeDefs.h"
#include "../Absyn.h"
#include "../Skeleton.h"
#include "Variable.h"
#include "../ParseError.h"
#include "../Info.h"
#include "Quadruple.h"
#include "BasicBlock.h"


class QuadrupleGenerator : public Skeleton {
public:
	struct Result {
		vector<BasicBlockPtr> blocks;
		vector<VariablePtr> vars;
	};
	explicit QuadrupleGenerator(shared_ptr<Scope> s) : lineno(0), scope(std::move(s)) {}
	Result compileFunction(FunctionInfoPtr f);
private:
	vector<VariablePtr> vars;
	vector<BasicBlockPtr> blocks;
	Quadruple::Op op;
	BasicBlockPtr block;
	VariablePtr lastVar;
	int lineno;
	shared_ptr<Scope> scope;
	shared_ptr<QLabel> labelTrue, labelFalse;

	VariablePtr evalLVal(Visitable *expr) {
		auto info = evalExpr(expr)->info;
		if (!info) {
			throw ParseError("LValue expected", expr);
		}
		info->loc = nullptr;
		return alloc(info);
	}

	VariablePtr evalJump(Visitable *expr, shared_ptr<QLabel> tLabel, shared_ptr<QLabel> fLabel) {
		if (!expr) throw runtime_error("No expr to eval");
		lineno = expr->lineno;
		lastVar = nullptr;
		swap(labelTrue, tLabel);
		swap(labelFalse ,fLabel);
		try {
			expr->accept(this);
		} catch (const ParseError &err) {
			throw ParseError(err, expr);
		}
		if (!lastVar) throw ParseError("No variable found", expr);
		swap(labelTrue, tLabel);
		swap(labelFalse ,fLabel);
		auto ret = lastVar;
		lastVar = nullptr;
		return ret;
	}

	inline VariablePtr evalExpr(Visitable *expr) {
		return evalJump(expr, nullptr, nullptr);
	}

	VariablePtr alloc(const VarInfoPtr& info) {
		if (info->loc) {
			return info->loc;
		}
		auto v = make_shared<Variable>(info);
		vars.emplace_back(v);
		if (!info->isInstanceVariable()) {
			info->loc = v;
		}
		return v;
	}

	template<typename... Args> VariablePtr alloc(Args... args) {
		auto v = make_shared<Variable>(args...);
		vars.emplace_back(v);
		return v;
	}

	template<typename T, typename... Args> shared_ptr<T> addQuad(Args... args) {
		return static_pointer_cast<T>(addQuad(make_shared<T>(args...)));
	}

	QuadruplePtr addQuad(QuadruplePtr quad) {
		quad->lineno = lineno;
		flushLastQuad();
		lastQuad = quad;
		return quad;
	}

	void addLastVarCondJump(const VariablePtr& l, Quadruple::Op op, const VariablePtr& r) {
		if (labelTrue && labelFalse) {
			addQuad<QJumpCond>(labelTrue, l, op, r);
			addQuad<QAssign>(lastVar, l, op, r);
			addQuad<QJump>(labelFalse);
			return;
		} else if (labelTrue) {
			addQuad<QJumpCond>(labelTrue, l, op, r);
			lastVar = alloc(0);
		} else if (labelFalse) {
			addQuad<QJumpCond>(labelFalse, l, op.neg(), r);
			lastVar = alloc(1);
		} else return;
		addQuad<QAssign>(lastVar, l, op, r);
	}

	QuadruplePtr lastQuad;

	BasicBlockPtr newBlock() {
		block = make_shared<BasicBlock>();
		block->phi->block = block;
		return block;
	}

	void flushLastQuad() {
		if (lastQuad) {
			block->quads.emplace_back(lastQuad);
			lastQuad->block = block;
			lastQuad = nullptr;
		}
	}

	BasicBlockPtr flushBasicBlock(bool append = false);

	void visitPlus(Plus *p) override { op = Quadruple::Op::Plus; }

	void visitMinus(Minus *p) override { op = Quadruple::Op::Minus; }

	VariableLayout captureEnv();

	void visitTimes(Times *p) override { op = Quadruple::Op::Mul; }

	void visitDiv(Div *p) override { op = Quadruple::Op::Div; }

	void visitMod(Mod *p) override { op = Quadruple::Op::Mod; }

	void visitLTH(LTH *p) override { op = Quadruple::Op::LT; }

	void visitLE(LE *p) override { op = Quadruple::Op::LE; }

	void visitEVar(EVar *p) override;

	void visitGTH(GTH *p) override { op = Quadruple::Op::GT; }

	void visitGE(GE *p) override { op = Quadruple::Op::GE; }

	void visitELitInt(ELitInt *p) override;

	void visitEQU(EQU *p) override { op = Quadruple::Op::EQ; }

	void visitNE(NE *p) override { op = Quadruple::Op::NEQ; }

	void visitELitTrue(ELitTrue *p) override;

	void visitELitFalse(ELitFalse *p) override;

	void visitEString(EString *p) override;

	void visitNullCast(NullCast *p) override;

	void visitNeg(Neg *p) override;

	void visitNot(Not *p) override;

	void visitEMul(EMul *p) override;

	void visitEAdd(EAdd *p) override;

	void visitERel(ERel *p) override;

	void visitEAnd(EAnd *p) override;

	void visitEOr(EOr *p) override;

	void visitEIndexAcc(EIndexAcc *p) override;

	void visitEClsMmbr(EClsMmbr *p) override;

	void visitEApp(EApp *p) override;

	void visitENewArray(ENewArray *p) override;

	void visitENewClass(ENewClass *p) override;

	void visitInit(Init *p) override;

	void visitNoInit(Init *p) override;

	void assign(Expr* lval, VariablePtr val);

	void visitAss(Ass *p) override;

	void visitIncr(Incr *p) override;

	void visitDecr(Decr *p) override;

	void visitRet(Ret *p) override;

	void visitVRet(VRet *p) override;

	void visitCond(Cond *p) override;

	void visitCondElse(CondElse *p) override;

	void compileCond(Expr *cond, Stmt *stmt1, Stmt *stmt2);

	void merge2Envs(VariableLayout *env1, const BasicBlockPtr& b1, VariableLayout *env2, const BasicBlockPtr& b2);

	void visitWhile(While *expr) override;

	void visitSExp(SExp *p) override;

	void visitForEach(ForEach *p) override;

	void visitBlk(Blk *p) override;

};


#endif //ZAD2_QUADRUPLEGENERATOR_H
